
;;;;
;;;;  SYSTEM MANIFESTOVANYCH TYPU
;;;;

;; vytvor element jazyka s manifestovanym typem
(define curry-make-elem
  (lambda (type-tag)
    (lambda (data)
      (cons type-tag data))))

;; vrat visacku s typem
(define get-type-tag car)

;; vrat data
(define get-data cdr)

;; test daneho typu
(define curry-scm-type
  (lambda (type)
    (lambda (elem)
      (equal? type (get-type-tag elem)))))

;;;;
;;;;  ZAKLADNI ELEMENTY JAZYKA SCHEME
;;;;

;;;; CISLA
;;;;;;;;;;;;;;;;;;;;

;; konstruktor cisla a test zda-li se jedna o cislo
(define make-number (curry-make-elem 'number))
(define scm-number? (curry-scm-type 'number))

;;;; SYMBOLY
;;;;;;;;;;;;;;;;;;;;

;; konstruktor symbolu a test zda-li se jedna o symbol
(define make-symbol (curry-make-elem 'symbol))
(define scm-named-symbol? (curry-scm-type 'symbol))

;; konstruktor generovaneho symbolu
(define make-generated-symbol
  (let ((make-physical-element (curry-make-elem 'generated-symbol)))
    (lambda ()
      (make-physical-element (cons #f #f)))))
(define scm-generated-symbol? (curry-scm-type 'generated-symbol))

;; test zda-li se jedna o symbol pojmenovany/generovany
(define scm-symbol?
  (lambda (elem)
    (or (scm-named-symbol? elem)
	(scm-generated-symbol? elem))))

;;;; TECKOVE PARY 
;;;;;;;;;;;;;;;;;;;;

;; konstruktor paru CONS
(define make-pair
  (let ((make-physical-pair (curry-make-elem 'pair)))
    (lambda (head tail)
      (make-physical-pair (mcons head tail)))))

;; test datoveho typu
(define scm-pair? (curry-scm-type 'pair))

;; selektor paru CAR
(define pair-car
  (lambda (pair)
    (if (scm-pair? pair)
	(mcar (get-data pair))
	(error "; Car: argument must be a pair"))))

;; selektor paru CDR
(define pair-cdr
  (lambda (pair)
    (if (scm-pair? pair)
	(mcdr (get-data pair))
	(error "; Cdr: argument must be a pair"))))

;; mutator paru SET-CAR!
(define pair-set-car!
  (lambda (pair value)
    (if (scm-pair? pair)
	(begin
	  (set-mcar! (get-data pair) value)
	  the-void-value)
	(error "; Set-car!: argument must be a pair"))))

;; mutator paru SET-CDR!
(define pair-set-cdr!
  (lambda (pair value)
    (if (scm-pair? pair)
	(begin
	  (set-mcdr! (get-data pair) value)
	  the-void-value)
	(error "; Set-cdr!: argument must be a pair"))))

;; obecny konvertor seznamu na scm-seznam a obracene
(define convert-list
  (lambda (a-null? a-car a-cdr b-cons b-nil f l)
    (if (a-null? l)
	b-nil
	(b-cons (f (a-car l))
		(convert-list a-null? a-car a-cdr b-cons b-nil f
			      (a-cdr l))))))

;;;; PROSTREDI
;;;;;;;;;;;;;;;;;;;;;;;;;

;; preved asociacni seznam na tabulku prostredi
(define assoc->env
  (lambda (l)
    (if (null? l)
	the-empty-list
	(make-pair (make-pair
		    (make-symbol (caar l))
		    (cdar l))
		   (assoc->env (cdr l))))))

;; konstruktor prostredi
(define make-env
  (let ((make-physical-env (curry-make-elem 'environment)))
    (lambda (pred table)
      (make-physical-env
       (cons pred (make-pair scm-false table))))))

;; test datoveho typu
(define scm-env? (curry-scm-type 'environment))

;; konstruktor globalniho prostredi
(define make-global-env
  (lambda (alist-table)
    (make-env scm-false (assoc->env alist-table))))

;; vrat tabulku
(define get-table
  (lambda (elem)
    (if (scm-env? elem)
	(cdr (get-data elem))
	(error "; Get-table: argument must be an environment"))))

;; vrat predka
(define get-pred
  (lambda (elem)
    (if (scm-env? elem)
	(car (get-data elem))
	(error "; Get-pred: argument must be an environment"))))

;; je globalni prostredi?
(define global?
  (lambda (elem)
    (and (scm-env? elem)
	 (equal? scm-false (get-pred elem)))))

;; hledani vazeb v asociacnim poli
(define scm-assoc
  (lambda (key alist)
    (cond ((scm-null? alist) scm-false)
	  ((eq? key (pair-car (pair-car alist))) (pair-car alist))
	  ((equal? key (pair-car (pair-car alist))) (pair-car alist))
	  (else (scm-assoc key (pair-cdr alist))))))

;; vyhledej vazbu v prostredi env, nebo vrat not-found
(define lookup-env
  (lambda (env symbol search-nonlocal? not-found)
    (let ((found (scm-assoc symbol (pair-cdr (get-table env)))))
      (cond ((not (equal? found scm-false)) found)
	    ((global? env) not-found)
	    ((not search-nonlocal?) not-found)
	    (else (lookup-env (get-pred env) symbol #t not-found))))))

;;;; PRIMITIVNI PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstruktor primitivni procedury a test zda-li se jedna o proceduru
(define make-primitive (curry-make-elem 'primitive))
(define scm-primitive? (curry-scm-type 'primitive))

;; vytvareni primitivnich procedur pomoci wrapperu
(define wrap-primitive
  (lambda (proc)
    (make-primitive
     (lambda arguments
       (expr->intern (apply proc (map get-data arguments)))))))

;; vytvareni predikatu pro testy typu pomoci wrapperu
(define wrap-predicate
  (lambda (type-pred)
    (make-primitive
     (lambda elems
       (if (apply type-pred elems)
	   scm-true
	   scm-false)))))

;;;; UZIVATELSKE PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstruktor procedury
(define make-procedure
  (let ((make-physical-procedure (curry-make-elem 'procedure)))
    (lambda (env args body)
      (make-physical-procedure (list env args body)))))

;; selektory
(define procedure-environment (lambda (proc) (car (get-data proc))))
(define procedure-arguments (lambda (proc) (cadr (get-data proc))))
(define procedure-body (lambda (proc) (caddr (get-data proc))))

;; test zda-li se jedna o proceduru
(define scm-user-procedure? (curry-scm-type 'procedure))

;; test zda-li se jedna o primitivni/uzivatelskou proceduru
(define scm-procedure?
  (lambda (elem)
    (or (scm-primitive? elem)
	(scm-user-procedure? elem))))

;;;; PRIMITIVNI SPECIALNI FORMY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstruktor primitivni spec. formy a test zda-li se jedna o spec. formu
(define make-specform (curry-make-elem 'specform))
(define scm-specform? (curry-scm-type 'specform))

;; test zda-li se jedna o primitivni/uzivatelskou formu
(define scm-form?
  (lambda (elem)
    (or (scm-specform? elem)
	(scm-macro? elem))))

;;;; MAKRA
;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstrukto makra
(define make-macro (curry-make-elem 'macro))
(define scm-macro? (curry-scm-type 'macro))

;; test zda-li se jedna o primitivni/uzivatelskou formu
(define scm-form?
  (lambda (elem)
    (or (scm-specform? elem)
	(scm-macro? elem))))

;;;; SPECIALNI ELEMENTY JAZYKA
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; pravdivostni hodnoty
(define scm-false ((curry-make-elem 'boolean) #f))
(define scm-true  ((curry-make-elem 'boolean) #t))
(define scm-boolean? (curry-scm-type 'boolean))

;; prazdny seznam
(define the-empty-list ((curry-make-elem 'empty-list) '()))
(define scm-null? (lambda (elem) (eq? elem the-empty-list)))

;; vrazdna hodnota
(define the-void-value ((curry-make-elem 'void) '()))
(define scm-void? (lambda (elem) (equal? elem the-void-value)))

;; nedefinovana hodnota
(define the-undefined-value ((curry-make-elem 'undefined) '()))
(define scm-undefined? (lambda (elem) (equal? elem the-undefined-value)))

;;;; GENERICKE PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; scm-eq? naprogramujeme pomoci interniho eq?
(define scm-eq?
  (lambda (elem-a elem-b)
    (eq? (get-data elem-a)
	 (get-data elem-b))))
