
;;;;
;;;;  POMOCNE PROCEDURY
;;;;

;; pro obecny typ volani apply sestavi seznam argumentu
(define apply-collect-arguments
  (lambda (args)
    (cond ((null? args) (error "; APPLY: argument missing"))
	  ((and (not (null? args)) (null? (cdr args)))
	   (scm-list->list (car args)))
	  (else (cons (car args) (apply-collect-arguments (cdr args)))))))

;;;;
;;;;  TOPLEVEL ENVIRONMENT
;;;;

;; vytvor prostredi, ktere je nejvys v hierarchii
(define scheme-toplevel-env
  (make-global-env
   `(
     ;; specialni forma IF
     (if . ,(make-specform
	     (lambda (env condition expr . alt-expr)
	       (let ((result (scm-eval condition env)))
		 (if (equal? result scm-false)
		     (if (null? alt-expr)
			 the-undefined-value
			 (scm-eval (car alt-expr) env))
		     (scm-eval expr env))))))

     ;; specialni forma AND
     (and . ,(make-specform
	      (lambda (env . exprs)
		(let and-eval ((exprs exprs))
		  (cond ((null? exprs) scm-true)
			((null? (cdr exprs)) (scm-eval (car exprs) env))
			(else (let ((result (scm-eval (car exprs) env)))
				(if (equal? result scm-false)
				    scm-false
				    (and-eval (cdr exprs))))))))))
     
     ;; specialni forma OR
     (or . ,(make-specform
	     (lambda (env . exprs)
	       (let or-eval ((exprs exprs))
		 (cond ((null? exprs) scm-false)
		       ((null? (cdr exprs)) (scm-eval (car exprs) env))
		       (else (let ((result (scm-eval (car exprs) env)))
			       (if (equal? result scm-false)
				   (or-eval (cdr exprs))
				   result))))))))

     ;; specialni forma LAMBDA
     (lambda . ,(make-specform
		 (lambda (env args . body)
		   (if (not (null? (cdr body)))
		       (make-procedure
			env args
			(make-pair (make-symbol 'begin)
				   (let iter ((ar body))
				     (if (null? (cdr ar))
					 (make-pair (car ar) the-empty-list)
					 (make-pair (car ar)
						    (iter (cdr ar)))))))
		       (make-procedure env args (car body))))))

     ;; specialni forma THE-ENVIRONMENT
     (the-environment . ,(make-specform (lambda (env) env)))

     ;; specialni forma QUOTE
     (quote . ,(make-specform (lambda (env elem) elem)))

     ;; specialni forma dyn-apply
     ;; prepisuje vyraz ve tvaru (dyn-apply proc a1 a2 ... (argumenty ...))
     ;; na vyraz (env-apply (the-environment) proc a1 a2 ... (argumenty ...))
     ;; a vyhodnoti jej v aktualnim prostredi
     (dyn-apply . ,(make-specform
		    (lambda (env proc . rest)
		      (scm-eval
		       (make-pair (make-symbol 'env-apply)
				  (make-pair
				   proc
				   (make-pair
				    env
				    (let iter ((ar rest))
				       (if (null? (cdr ar))
					   (make-pair (car ar) the-empty-list)
					   (make-pair (car ar)
						      (iter (cdr ar))))))))
		       env))))

     ;; specialni forma BEGIN
     (begin . ,(make-specform
		(lambda (env . body)
		  (let iter ((body body))
		    (cond ((null? body) the-undefined-value)
			  ((null? (cdr body)) (scm-eval (car body) env))
			  (else (begin 
				  (scm-eval (car body) env)
				  (iter (cdr body)))))))))

     ;; specialni forma DEFINE
     (define . ,(make-specform
		 (lambda (env symbol value)
		   (let ((value (scm-eval value env))
			 (result (lookup-env env symbol #f #f)))
		     (if result
			 (pair-set-cdr! result value)
			 (pair-set-cdr!
			  (get-table env)
			  (make-pair (make-pair symbol value)
				     (pair-cdr (get-table env)))))))))
     
     ;; specialni forma SET!
     (set! . ,(make-specform
	       (lambda (env symbol value)
		 (pair-set-cdr! (lookup-env env symbol #t #f)
				(scm-eval value env)))))

     ;; konstanty
     (pi . ,(make-number (* 4 (atan 1))))

     ;; procedury pro praci s cisly (zakomentovane procedury jsou v R5RS, ale
     ;; nejsou implementovany ve vsech dostupnych interpretech Scheme)
     (* . ,(wrap-primitive *))
     (+ . ,(wrap-primitive +))
     (- . ,(wrap-primitive -))
     (/ . ,(wrap-primitive /))
     (< . ,(wrap-primitive <))
     (<= . ,(wrap-primitive <=))
     (= . ,(wrap-primitive =))
     (> . ,(wrap-primitive >))
     (>= . ,(wrap-primitive >=))
     (abs . ,(wrap-primitive abs))
     (acos . ,(wrap-primitive acos))
;     (angle . ,(wrap-primitive angle))
     (asin . ,(wrap-primitive asin))
     (atan . ,(wrap-primitive atan))
     (ceiling . ,(wrap-primitive ceiling))
     (cos . ,(wrap-primitive cos))
;     (denominator . ,(wrap-primitive denominator))
     (even? . ,(wrap-primitive even?))
     (exact->inexact . ,(wrap-primitive exact->inexact))
     (exact? . ,(wrap-primitive exact?))
     (exp . ,(wrap-primitive exp))
     (expt . ,(wrap-primitive expt))
     (floor . ,(wrap-primitive floor))
     (gcd . ,(wrap-primitive gcd))
;     (imag-part . ,(wrap-primitive imag-part))
     (inexact->exact . ,(wrap-primitive inexact->exact))
     (inexact? . ,(wrap-primitive inexact?))
     (lcm . ,(wrap-primitive lcm))
     (log . ,(wrap-primitive log))
;     (magnitude . ,(wrap-primitive magnitude))
;     (make-polar . ,(wrap-primitive make-polar))
;     (make-rectangular . ,(wrap-primitive make-rectangular))
     (max . ,(wrap-primitive max))
     (min . ,(wrap-primitive min))
     (modulo . ,(wrap-primitive modulo))
     (negative? . ,(wrap-primitive negative?))
;     (numerator . ,(wrap-primitive numerator))
     (odd? . ,(wrap-primitive odd?))
     (positive? . ,(wrap-primitive positive?))
     (quotient . ,(wrap-primitive quotient))
;     (rationalize . ,(wrap-primitive rationalize))
;     (real-part . ,(wrap-primitive real-part))
     (remainder . ,(wrap-primitive remainder))
     (round . ,(wrap-primitive round))
     (sin . ,(wrap-primitive sin))
     (sqrt . ,(wrap-primitive sqrt))
     (tan . ,(wrap-primitive tan))
     (truncate . ,(wrap-primitive truncate))
     (zero? . ,(wrap-primitive zero?))

     ;; prace s pary
     (cons . ,(make-primitive make-pair))
     (car . ,(make-primitive pair-car))
     (cdr . ,(make-primitive pair-cdr))
     (set-car! . ,(make-primitive pair-set-car!))
     (set-cdr! . ,(make-primitive pair-set-cdr!))

     ;; negace
     (not . ,(make-primitive
	      (lambda (elem)
		(if (equal? elem scm-false)
		    scm-true
		    scm-false))))

     ;; predikaty
     (boolean? . ,(wrap-predicate scm-boolean?))
     (form? . ,(wrap-predicate scm-form?))
     (null? . ,(wrap-predicate scm-null?))
     (number? . ,(wrap-predicate scm-number?))
     (pair? . ,(wrap-predicate scm-pair?))
     (primitive? . ,(wrap-predicate scm-primitive?))
     (procedure? . ,(wrap-predicate scm-procedure?))
     (specform? . ,(wrap-predicate scm-specform?))
     (symbol? . ,(wrap-predicate scm-symbol?))
     (undefined? . ,(wrap-predicate scm-undefined?))
     (user-procedure? . ,(wrap-predicate scm-user-procedure?))
     (void? . ,(wrap-predicate scm-void?))
     (eq? . ,(wrap-predicate scm-eq?))
     
     ;; sektoru uzivatelskych procedur
     (procedure-environment . ,(make-primitive procedure-environment))
     (procedure-arguments . ,(make-primitive procedure-arguments))
     (procedure-body . ,(make-primitive procedure-body))

     ;; procedura environment-parent (vrat lexikalniho predka daneho prostredi)
     (environment-parent . ,(make-primitive get-pred))

     ;; procedura environment->list (vrat tabulku s vazbami v danem prosredi)
     (environment->list . ,(make-primitive
			    (lambda (elem)
			      (if (equal? elem scm-false)
				  scm-false
				  (get-table elem)))))

     ;; procedura eval (dvou argumentu, druhy je vzdy povinny)
     (eval . ,(make-primitive
	       (lambda (elem env)
		 (scm-eval elem env))))

     ;; procedura apply
     (apply . ,(make-primitive
		(lambda (proc . rest)
		  (scm-apply proc (apply-collect-arguments rest)))))

     ;; procedura macro + vracec transformatoru
     (macro . ,(make-primitive make-macro))
     (macro-transformer . ,(make-primitive get-data))

     ;; generovane symboly
     (gensym . ,(make-primitive make-generated-symbol))

     ;; procedura apply, ktere lze predat prostredi predka
     (env-apply . ,(make-primitive
		    (lambda (proc env . rest)
		      (scm-env-apply proc
				     env
				     (apply-collect-arguments rest)))))

     ;; procedura lookup-env
     (lookup-env . ,(make-primitive
		     (lambda (env symbol . nonlocal)
		       (lookup-env env symbol
				   (or (null? nonlocal)
				       (equal? (car nonlocal) scm-true))
				   scm-false))))

     )))
